/* lmpprint.c
 *
 * Copyright (C) 1999-2000 Bernhard Pietsch <bjtp@gmx.net>
 *
 * This file is part of LMP library.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "lmp.h"
#include "lmpmath.h"
#include "lmpinput.h"
#include <stdio.h>
#include <string.h>
#include <malloc.h>

#ifndef HAVE_ASPRINTF

#warning No asprintf function !
#include <stdarg.h>
/* THIS IS A HACK */

void asprintf (char ** in, const char * format, ...) {
  va_list args;

  *in = malloc (4097);
  va_start (args, format);
  vsprintf (*in, format, args);
  va_end (args);
}

#endif

void lmp_print_tree (const char * lead, const char * add, lmp_node_t * root) {
  lmp_node_t * p;
  char * s;
  int i;

  s = malloc (strlen (lead) + strlen (add) + 1);
  strcpy (s, lead);
  strcat (s, add);
  p = root;
  if (p) {
    switch (p->what) {
    case OP:
      printf ("%s%s\n", lead, lmp_infix_op2str (p->data.operation.op));
      for (i = 0; i < p->data.operation.argc; i++) {
        lmp_print_tree (s, add, p->data.operation.argv[i]);
      }
      break;
    case CONTROL:
      printf ("%s%s\n", lead, lmp_infix_ctrl2str (p->data.control.op));
      for (i = 0; i < p->data.control.argc; i++) {
        lmp_print_tree (s, add, p->data.control.argv[i]);
      }
      break;
    case VAL:
      printf ("%s{%" FREAL ",%" FREAL "}\n", lead, CREAL (p->data.value), CIMAG (p->data.value));
      break;
    case VAR:
      printf ("%s%s\n", lead, p->data.variable);
      break;
    }
  }

  free (s);
}

void lmp_print_tree_pc (const char * lead, const char * add, lmp_node_t * root) {
  lmp_node_t * p;
  char * s;
  int i;

  s = malloc (strlen (lead) + strlen (add) + 1);
  strcpy (s, lead);
  strcat (s, add);
  p = root;
  if (p) {
    switch (p->what) {
    case OP:
      printf ("%s%s\n", lead, lmp_infix_op2str (p->data.operation.op));
      for (i = 0; i < p->data.operation.argc; i++) {
        lmp_print_tree_pc (s, add, p->data.operation.argv[i]);
      }
      break;
    case CONTROL:
      lmp_print_tree_pc (s, add, lmp_proc_control (root));
      break;
    case VAL:
      printf ("%s{%" FREAL ",%" FREAL "}\n", lead, CREAL (p->data.value), CIMAG (p->data.value));
      break;
    case VAR:
      printf ("%s%s\n", lead, p->data.variable);
      break;
    }
  }

  free (s);
}

void lmp_print_number (lmp_double_t val) {
  static char format [] = "%xxx.xxxxxx";

#ifdef FCVT
  char * nump, *num;
  int l, i;
#else
  lmp_double_t v;
#endif

  if (FINITE(val)) {
#ifdef FCVT
    int decpt, neg;
#endif

#ifndef FCVT
#warning "no FCVT function found"
//    MODF (val, &v);
    v = val;

    sprintf (format, "%%.%dLg", (int) (LOG10 (ABS (v) + 1) + 7.0));
    printf (format, val);

#else

    nump = FCVT (val, 6, &decpt, &neg);
    l = strlen (nump);
    num = malloc (l + 1);
    strcpy (num, nump);

    for (i = l-1; i >= 0; i--)
      if (num [i] != '0')
	break;
      else if (i >= decpt)
	num [i] = '\0';

    if (i < 0) {
      free (num);
      fprintf (stdout, "0");
      return;
    }

    if (neg)
      fputc ('-', stdout);
    if (decpt <= 0)
      fputc ('0', stdout);
    else {
      sprintf (format, "%%.%ds", decpt);
      fprintf (stdout, format, num);
    }
    if (num [abs (decpt)] != '\0')
      fputc ('.', stdout);
    if (decpt <= 0) {
      for (i = 0; i > decpt; i--)
	fputc ('0', stdout);
    }
    printf ("%s", num + (decpt > 0 ? decpt : 0));
    free (num);
#endif
  }
  else
    printf ("%" FREAL, val);
}

char * lmp_print_number_to_string (lmp_double_t val) {
  static char format [] = "%xxx.xxxxxx";
  char * output, * tmp;
  int ci;

#ifdef FCVT
  char * nump, *num;
  int l, i;
#else
  lmp_double_t v;
#endif

  if (FINITE(val)) {
#ifdef FCVT
    int decpt, neg;
#endif

#ifndef FCVT
#warning "no FCVT function found"

//    MODF (val, &v);

    v = val;

    sprintf (format, "%%.%dLg", (int) (LOG10 (ABS (v) + 1) + 7.0));
    asprintf (&output, format, val);

//    asprintf (&output, "%" FREAL, val);
	
#else

    nump = FCVT (val, 6, &decpt, &neg);
    l = strlen (nump);
    num = malloc (l + 1);
    strcpy (num, nump);

    for (i = l-1; i >= 0; i--)
      if (num [i] != '0')
	break;
      else if (i >= decpt)
	num [i] = '\0';

    if (i < 0) {
      free (num);
      output = malloc (2 * sizeof (char));
      strcpy (output, "0");
      return output;
    }

    output = malloc (3 * sizeof (char));
    ci = 0;
    if (neg)
      output [ci++] = '-';
    if (decpt <= 0)
      output [ci++] = '0';
    else {
      output [ci] = '\0';
      sprintf (format, "%%.%ds", decpt);
      asprintf (&tmp, format, num);
      output = realloc (output, ci + 1 + strlen (tmp));
      strcat (output, tmp);

      ci += strlen (tmp);
      free (tmp);
    }
    if (num [abs (decpt)] != '\0') {
      output = realloc (output, ci + 2);
      output [ci++] = '.';
    }
    if (decpt <= 0) {
      for (i = 0; i > decpt; i--) {
	output = realloc (output, ci + 2);
	output[ci++] = '0';
      }
    }
    asprintf (&tmp, "%s", num + (decpt > 0 ? decpt : 0));
    free (num);

    output [ci] = '\0';
    strcat (output, tmp);

    free (tmp);
#endif

  }
  else
    asprintf (&output, "%" FREAL, val);

  return output;
}
