/* lmpscan.c
 *
 * Copyright (C) 1999-2000 Bernhard Pietsch <bjtp@gmx.net>
 *
 * This file is part of LMP library.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "lmp.h"
#include "lmpinput.h"

/* Parse string to an array of strings.
 * lmp_parse_string() parses string. argc is set to the
 * number of elements lmp_parse_string() will found.
 *
 * This function may be used before lmp_eval().
 * returns the strings vector.
 */
char ** lmp_iss_by_space (string,argc)
const char * string; /* string to be parsed */
int * argc;          /* returns the number of elements */
{
  char ** argv, * sp, * tok;

  argv = NULL;
  *argc = 0;

  if (!string)
    return NULL;

  sp = malloc (strlen (string) + 1);
  strcpy (sp, string);

  while ((tok = GETTOKEN (sp, " ")) != NULL) {
    (*argc)++;
    argv = realloc (argv, *argc * sizeof (char **));
    argv [*argc-1] = malloc (strlen (tok) + 1);
    strcpy (argv [*argc-1], tok);
  }
  free (sp);

  return argv;
}

/* Parse string to an array by scanning
 * returns the string vector
 */
char ** lmp_iss_by_elements (s,c) 
char * s;
int * c;
{
  enum { NOTHING, DIGIT, UNARY_MINUS, MANT, SPACE, OPENSYMBOL, CLOSESYMBOL, SYMBOL, ALPHA, PLUS, MINUS, ASTERISK, AND, OR, EQUAL, LOWER, GREATER, NOT } state;
  char ** res, * last;
  int len, i, str;

  str = len = 0;
  state = NOTHING;
  res = malloc ((str+1)*sizeof (char *));
  last = s;
  for (i = 0; i < strlen (s); i++) {
    if ((s[i] >= '0' && s[i] <= '9') || (s[i] == '.')) { // Digit
      if ((state != DIGIT) && (state != MANT) && (state != UNARY_MINUS)) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
      }
      state = DIGIT;
      len++;
    }
    else if ((state == DIGIT) && ((s[i] == 'e') || (s[i] == 'E'))) {
      len++;
      state = MANT;
    }
    else if ((s[i] == ' ') || (s[i] == '\t') || (s[i] == '\n')) { // Space
      if (state != NOTHING) {
	res[str] = malloc (len + 1);
	strncpy (res[str], last, len);
	res[str++][len] = '\0';
	res = realloc (res, (str+1)*sizeof (char *));
      }
      last = s+i+1;
      state = NOTHING;
      len = 0;
    }
    else if ((s[i] == '(') || (s[i] == '{')) { // OpenSymbol
      if (state != NOTHING) {
	res[str] = malloc (len + 1);
	strncpy (res[str], last, len);
	res[str++][len] = '\0';
	res = realloc (res, (str+1)*sizeof (char *));
	last = s+i;
      }
      len = 1;
      state = OPENSYMBOL;
    }
    else if ((s[i] == ')') || (s[i] == '}')) { // CloseSymbol
      if (state != NOTHING) {
	res[str] = malloc (len + 1);
	strncpy (res[str], last, len);
	res[str++][len] = '\0';
	res = realloc (res, (str+1)*sizeof (char *));
	last = s+i;
      }
      len = 1;
      state = CLOSESYMBOL;
    }
    else if ((s[i] == '}') || (s[i] == ',') || (s[i] == '/') || (s[i] == '~') || (s[i] == '^')) { // Symbol
      if (state != NOTHING) {
	res[str] = malloc (len + 1);
	strncpy (res[str], last, len);
	res[str++][len] = '\0';
	res = realloc (res, (str+1)*sizeof (char *));
	last = s+i;
      }
      len = 1;
      state = SYMBOL;
    }
    else switch (s[i]) {
    case '*':
      if (state != ASTERISK) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = ASTERISK;
      }
      len++;
      break;
    case '+':
      if (state != PLUS) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = PLUS;
      }
      len++;
      break;      
    case '|':
      if (state != OR) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = OR;
      }
      len++;
      break;
    case '&':
      if (state != AND) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = AND;
      }
      len++;
      break;      
    case '-':
      if (state == MANT) {
	state = DIGIT; // little bit shit, but works
      }
      else if ((state == NOTHING) || (state == OPENSYMBOL) || (state == SYMBOL) || (state == SPACE)|| (state == EQUAL) || (state == NOT) || (state == AND) || (state == OR) || (state == GREATER) || (state == LOWER)) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = UNARY_MINUS;
      }
      else if (state != MINUS) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = MINUS;
      }
      len++;
      break;      
    case '=':
      if ((state != EQUAL) && (state != PLUS) && (state != MINUS) && (state != ASTERISK) && (state != AND) && (state != OR) && (state != LOWER) && (state != GREATER) && (state != NOT)) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = EQUAL;
      }
      len++;
      break;      
    case '!':
      if (state != NOT) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = NOT;
      }
      len++;
      break;      
    default: { // Alphas
	if (state != ALPHA) {
	  if (state != NOTHING) {
	    res[str] = malloc (len + 1);
	    strncpy (res[str], last, len);
	    res[str++][len] = '\0';
	    res = realloc (res, (str+1)*sizeof (char *));
	    last = s+i;
	  }
	  len = 0;
	  state = ALPHA;
	}
	len++;
      }
      break;
    }
  }

  if (state != NOTHING) {
    res[str] = malloc (len + 1);
    strncpy (res[str], last, len);
    res[str++][len] = '\0';
    res = realloc (res, (str+1)*sizeof (char *));
    last = s+i;
    len = 0;
  }

  res [str] = NULL;
  *c = str;

  return res;
}
